import os, sys
from dataclasses import dataclass

import xarray as xr
import numpy as np
import scipy.stats as stats

import tinysunset as ts
from tinysunset.scalers.base import ScalerConfig, BaseScaler

class GammaScaler(BaseScaler):
    def __init__(self, config: ScalerConfig):
        super().__init__(config)

    def _fit_gamma(self, values):
        #try:
        #    params = stats.gamma.fit(x, floc=0)
        #    return params[0:2]
        #except:
        #    return (np.nan, np.nan)
       #def get_gamma_params(values):
        means = np.nanmean(values, axis=0)
        log_means = np.log(means)
        logs = np.log(values)
        mean_logs = np.nanmean(logs, axis=0)
        a = log_means - mean_logs
        alphas = (1 + np.sqrt(1 + 4 * a / 3)) / (4 * a)
        betas = means / alphas

        return (alphas, betas)
        
    def _fit(self, da: xr.DataArray):

        mean, var = xr.apply_ufunc(
            lambda x: self._fit_gamma(x.flatten()),
            da,
            input_core_dims=[[self.time_dim, self.ensemble_dim]],
            output_core_dims=[[],[]],
            vectorize=True,
            output_dtypes=[None, None],
            dask='parallelized',
        )

        da = xr.apply_ufunc(
            lambda da, mean, var: stats.norm.ppf(stats.gamma.cdf(da, a=mean, scale=var)),
            #lambda data, params: stats.norm.ppf(stats.gamma.cdf(data, **params)),
            da, mean, var,
            input_core_dims=[[self.time_dim, self.ensemble_dim], [], []],
            output_core_dims=[[self.time_dim, self.ensemble_dim]],
            vectorize=True,
            dask='parallelized',
            output_dtypes=[None]
        )
        
        return mean, var
    
    def fit(self, da: xr.DataArray):
        
        
        da = da.where(da > 0, 0)
        da = da.sortby(self.time_dim)

        _func = {
            #"ens_mean_loess_retro": lambda x: ts.tools.loess_ts_retro(x, window=self.loess_window, degree=self.loess_degree),
            "ens_mean_loess": lambda x: ts.tools.loess_ts(x, window=self.loess_window, degree=self.loess_degree),
            "ens_mean": lambda x: ts.tools.loess_ts(x, window=self.loess_window, degree=self.loess_degree),
        }

        if self.groupby_type is not None:

            self.mean, self.var = self._fit(da.groupby(self.groupby_type))

            if self.clim_type in ["ens_mean_loess", "ens_mean"]:

                self.trend = self._transform(
                    da.groupby(self.groupby_type)
                    ).groupby(self.groupby_type).mean(dim=self.ensemble_dim, skipna=True)
                

                self.trend = xr.apply_ufunc(
                    _func[self.clim_type],
                    self.trend,
                    input_core_dims = [[self.time_dim]],
                    output_core_dims=[[self.time_dim]],
                    vectorize=True,
                    dask='parallelized',
                    output_dtypes=[float]
                    )

            else:

                self.trend = None
        else:

            self.mean, self.var = self._fit(da)
            #if self.clim_type == "ens_mean":
            #    self.trend = self._transform(da).mean(dim=self.ensemble_dim, skipna=True)
            if self.clim_type in [ "ens_mean", "ens_mean_loess"]:

               self.trend = self._transform(da).mean(dim=self.ensemble_dim, skipna=True)
               self.trend = xr.apply_ufunc(
                    _func[self.clim_type],
                    self.trend,
                    input_core_dims = [[self.time_dim]],
                    output_core_dims=[[self.time_dim]],
                    vectorize=True,
                    dask='parallelized',
                    output_dtypes=[float]
                    )
            else:
                self.trend = None
            
    
    def transform(self, da: xr.DataArray):
        
        da = da.where(da > 0, 0)

        da = da.sortby(self.time_dim)
        
        if self.groupby_type is not None:
            da = self._transform(da.groupby(self.groupby_type))
            if self.trend is not None:
                da = da.groupby(self.groupby_type) - self.trend
        else:
            da = self._transform(da) 
            if self.trend is not None:
                da = da - self.trend
        
        return da
        
    def inverse_transform(self, da: xr.DataArray):
        
        if self.groupby_type is not None:

            if self.trend is not None:
                da = da.groupby(self.groupby_type) + self.trend
            return self._inverse_transform(da.groupby(self.groupby_type))
        else:
            if self.trend is not None:
                da = da + self.trend
            
            return self._inverse_transform(da)
        
    def fit_transform(self, da: xr.DataArray):
        
        da = da.where(da > 0, 0)
        
        self.fit(da)
        
        return self.transform(da)
    
    def _transform(self, da: xr.DataArray):

        return xr.apply_ufunc(
            lambda data, mean, var: stats.norm.ppf(stats.gamma.cdf(data, a=mean, scale=var)),
            #lambda data, params: stats.norm.ppf(stats.gamma.cdf(data, **params)),
            da, self.mean, self.var,
            input_core_dims=[[self.time_dim, self.ensemble_dim], [], []],
            output_core_dims=[[self.time_dim, self.ensemble_dim]],
            vectorize=True,
            dask='parallelized',
            output_dtypes=[None]
        ) #- self.trend if self.trend is not None else 0

    def _inverse_transform(self, da: xr.DataArray):
        
        da = xr.apply_ufunc(
            lambda data, mean, var: stats.gamma.ppf(stats.norm.cdf(data), a=mean, scale=var),
            da, self.mean, self.var,
            input_core_dims=[[self.time_dim, self.ensemble_dim], [], []],
            output_core_dims=[[self.time_dim, self.ensemble_dim]],
            vectorize=True,
            dask='parallelized',
            output_dtypes=[float]
        )
        
        return da