

import os, sys
from dataclasses import dataclass

import xarray as xr
import numpy as np

import tinysunset as ts
from tinysunset.scalers.base import ScalerConfig, BaseScaler

class RobustScaler(BaseScaler):
    """
    RobustScaler that uses median and interquartile range (IQR) instead of mean and std.
    This makes the scaling more robust to outliers in the data.
    
    The transformation is: (X - median) / (IQR * scale_factor)
    Where IQR = Q3 - Q1 (75th percentile - 25th percentile)
    
    The default scale_factor of 1.349 makes the scaling comparable to StandardScaler for normally distributed data.
    """

    def __init__(
        self, 
        scaler_config: ScalerConfig
        ):
        
        super().__init__(scaler_config)

        self.alpha = scaler_config.alpha
        self.scale_factor = 1.349  # Makes IQR scaling comparable to std for normal distribution

    def _fit(self, da: xr.DataArray):
        
        assert self.ensemble_dim in list(da.dims), f"{self.ensemble_dim} not in ds dims: {da.dims}"
        assert self.time_dim in list(da.dims), f"{self.time_dim} not in ds dims: {da.dims}"

        da = da.sortby(self.time_dim)
    
        if self.clim_type == "clim":
            # Use median instead of mean for central tendency
            assert self.clim_ref_years is not None, "clim_ref_years must be specified"
            _selection = {self.time_dim: slice(str(self.clim_ref_years[0]), str(self.clim_ref_years[-1]))}
            _data_slice = da.sel(_selection)
            
            if self.groupby_type is not None:
                self.median = _data_slice.groupby(self.groupby_type).median([self.time_dim, self.ensemble_dim])
            else:
                self.median = _data_slice.median([self.time_dim, self.ensemble_dim])

        else:
            if self.groupby_type is not None:
                # Get median across ensemble members
                self.median = da.groupby(self.groupby_type).median([self.ensemble_dim])
            else:
                self.median = da.median([self.ensemble_dim])

            if self.clim_type == "ens_mean":
                # No additional processing - keep the ensemble median time series
                pass
                
            elif self.clim_type == "ens_mean_loess":
                # Apply LOESS smoothing to the ensemble median time series
                self.median = xr.apply_ufunc(
                    lambda x: ts.tools.loess_ts(x, window=self.loess_window, degree=self.loess_degree),
                    self.median,
                    input_core_dims=[[self.time_dim]],
                    output_core_dims=[[self.time_dim]],
                    vectorize=True,
                    dask='parallelized',
                    output_dtypes=[float]
                )
            
            elif self.clim_type == "ens_mean_loess_retro":
                # Apply retrospective LOESS smoothing
                self.median = xr.apply_ufunc(
                    lambda x: ts.tools.loess_ts_retro(x, window=self.loess_window, degree=self.loess_degree),
                    self.median,
                    input_core_dims=[[self.time_dim]],
                    output_core_dims=[[self.time_dim]],
                    vectorize=True,
                    dask='parallelized',
                    output_dtypes=[float]
                )

            elif self.clim_type == "exp_smoothing":
                # Apply exponential smoothing
                self.median = xr.apply_ufunc(
                    lambda x: ts.tools.exp_smoothing_ts(x, alpha=self.alpha),
                    self.median,
                    input_core_dims=[[self.time_dim]],
                    output_core_dims=[[self.time_dim]],
                    vectorize=True,
                    dask='parallelized',
                    output_dtypes=[float]
                )

            elif self.clim_type == "exp_smoothing_retro":
                # Apply retrospective exponential smoothing
                self.median = xr.apply_ufunc(
                    lambda x: ts.tools.exp_smoothing_ts_retro(x, alpha=self.alpha),
                    self.median,
                    input_core_dims=[[self.time_dim]],
                    output_core_dims=[[self.time_dim]],
                    vectorize=True,
                    dask='parallelized',
                    output_dtypes=[float]
                )
            
            elif self.clim_type == "retroactive":
                # Apply rolling median for retrospective analysis
                self.median = self.median.shift({self.time_dim: 1}).rolling({self.time_dim: self.retro_window}).median()
            else:
                raise ValueError(f"clim_type {self.clim_type} not recognized")

        # Calculate IQR instead of std for dispersion
        if self.clim_type == "retroactive":
            # For retroactive mode, calculate rolling IQR
            q75 = da.shift({self.time_dim: 1}).rolling({self.time_dim: self.retro_window}).quantile(0.75, dim=[self.time_dim, self.ensemble_dim])
            q25 = da.shift({self.time_dim: 1}).rolling({self.time_dim: self.retro_window}).quantile(0.25, dim=[self.time_dim, self.ensemble_dim])
            self.iqr = (q75 - q25) / self.scale_factor
        else:
            if self.std_ref_years is not None:
                # Calculate IQR based on reference years
                _selection = {self.time_dim: slice(str(self.std_ref_years[0]), str(self.std_ref_years[-1]))}
                _data_slice = da.sel(_selection)
                
                if self.groupby_type is not None:
                    # Center the data around the median before calculating IQR
                    _centered = _data_slice.groupby(self.groupby_type) - self.median
                    q75 = _centered.groupby(self.groupby_type).quantile(0.75, dim=[self.time_dim, self.ensemble_dim])
                    q25 = _centered.groupby(self.groupby_type).quantile(0.25, dim=[self.time_dim, self.ensemble_dim])
                else:
                    # Center the data around the median before calculating IQR
                    _centered = _data_slice - self.median
                    q75 = _centered.quantile(0.75, dim=[self.time_dim, self.ensemble_dim])
                    q25 = _centered.quantile(0.25, dim=[self.time_dim, self.ensemble_dim])
                
                self.iqr = (q75 - q25) / self.scale_factor
            else:
                # No dispersion scaling
                self.iqr = 1  # = No standardization
                
    def fit(self, da: xr.DataArray):
        """
        Fit the scaler to the data
        """
        self._fit(da)

    def transform(self, da: xr.DataArray):
        """
        Transform the data using the formula: (X - median) / (IQR / scale_factor)
        """
        da = da.sortby(self.time_dim)

        if self.groupby_type is not None:
            return (da.groupby(self.groupby_type) - self.median).groupby(self.groupby_type) / self.iqr
        else:
            return (da - self.median) / self.iqr
            
    def fit_transform(self, da: xr.DataArray):
        """
        Fit the scaler to the data and transform it in one operation
        """
        self.fit(da)
        return self.transform(da)

    def inverse_transform(self, da: xr.DataArray):
        """
        Inverse transform the data: X * IQR + median
        """
        da = da.sortby(self.time_dim)

        if self.groupby_type is not None:
            return da.groupby(self.groupby_type) * self.iqr + self.median
        else:
            return da * self.iqr + self.median
        
        
class RobustScalerPR(RobustScaler):
    """
    RobustScaler specifically for precipitation data.
    It uses a different scale factor for precipitation data to account for its distribution.
    """
    
    def __init__(
        self, 
        scaler_config: ScalerConfig
        ):
        
        super().__init__(scaler_config)
        self.scale_factor = 0 # the std won't be used for pr, only the median
        
    def fit(self, da: xr.DataArray):
        
        da = self._clip_negatives_to_zero(da)
        
        self._fit(da) 
        
    def transform(self, da: xr.DataArray):
        """
        Transform the data using the formula: (X - median) / (IQR / scale_factor)
        """
        
        da = self._clip_negatives_to_zero(da)
        
        da = da.sortby(self.time_dim)

        if self.groupby_type is not None:
            da = (da.groupby(self.groupby_type)/self.median)
        else:
            da = da/self.median

        return da
        
            
    def fit_transform(self, da: xr.DataArray):
        """
        Fit the scaler to the data and transform it in one operation
        """
        self.fit(self._clip_negatives_to_zero(da))
        return self.transform(da)

    def inverse_transform(self, da: xr.DataArray):
        """
        Inverse transform the data: X * IQR + median
        """
        da = da.sortby(self.time_dim)

        if self.groupby_type is not None:
            return da.groupby(self.groupby_type) * self.median
        else:
            return da * self.median
        
    def _clip_negatives_to_zero(self, da: xr.DataArray):
        """
        Clip negative values to zero
        """
        return da.where(da > 0, 0)