
import xarray as xr
import numpy as np
import xskillscore as xs
from scipy.stats import ks_2samp

import matplotlib.pyplot as plt
import cartopy.crs as ccrs
from matplotlib.colors import LinearSegmentedColormap

from tinysunset.plots import plot_array_to_map
from tinysunset.tools import _xbootstrap

def _skill_fun(metric, ref_metric):
    return 1 - (metric / ref_metric)

_SKILL_scores = ["crpsgs", "crpss", "rpss", "bss", "rmses", "maes", "mses"]

_BOOTSTRAP_METRICS_random = [
    "ensmeanpearsoncorr", "ensmedianpearsoncorr", "ensmeanspearmancorr", "ensmedianspearmancorr"
]
    
_BOOTSTRAP_METRICS_zero = [
    "crpss", "crpsgs", "rpss", "bss", "rmses", "maes", "mses", "meanbiass", "medians"
]
    
_BOOTSTRAP_METRICS = _BOOTSTRAP_METRICS_random + _BOOTSTRAP_METRICS_zero



def _compute_single_metric(metric, fcst_data, obs_data, METRICS,
                           suffix="", ref_fcst=None, 
                           time_dim="time", bootstrap=False, bootstrap_nsim=500, p_val_thrs=0.05
                           ):
    
    
    """Helper function to compute a single metric with optional suffix"""
    
    if metric in _SKILL_scores:
        if ref_fcst is not None:
            # drop just the last letter
            _metric_score = metric[:-1]
            _fun = lambda fcst, obs, ref_fcst: _skill_fun(METRICS[_metric_score](fcst, obs), METRICS[_metric_score](ref_fcst, obs))
            _arrays = [fcst_data, obs_data, ref_fcst]
        else:
            raise ValueError(f"Reference forecast is required to compute skill score for metric '{metric}'.")
    else:
        _fun = METRICS[metric]
        _arrays = [fcst_data, obs_data]

    metric_name = f"{metric}{suffix}"
    results = []

    if bootstrap and metric in _BOOTSTRAP_METRICS:
        test = _xbootstrap(_fun, time_dim, bootstrap_nsim, *_arrays)
        results.append(test.median("n_sim").rename(metric_name))
        
        if metric in _BOOTSTRAP_METRICS_random:
            random_test = _xbootstrap(
                _fun, time_dim, bootstrap_nsim,
                fcst_data.copy()*np.random.rand(*fcst_data.shape)/fcst_data, obs_data
            )
            
            p_val, stats = xr.apply_ufunc(
                ks_2samp, random_test, test,
                input_core_dims=[["n_sim"], ["n_sim"]],
                output_core_dims=[[],[]],
                vectorize=True, dask="parallelized",
                output_dtypes=[float, float]
            )
            results.append((1 - p_val).rename(f"{metric_name}_pval"))
            
        elif metric in _BOOTSTRAP_METRICS_zero:
            p_val = test.quantile(p_val_thrs, dim="n_sim")
            results.append(p_val.rename(f"{metric_name}_pval"))
    else:
        results.append(_fun(*_arrays).rename(metric_name))
                          
    return results


def comp_metrics(
    fcst: xr.DataArray, 
    obs: xr.DataArray, 
    metrics: list,
    ref_fcst: xr.DataArray = None,
    include_ref_metrics: bool = False,  # NEW PARAMETER
    time_dim: str = "time",
    ensemble_dim: str = "ensemble",
    rps_thresholds: np.ndarray = np.array([-0.431, 0.431]),
    bs_threshold: float = -1,
    bootstrap: bool = False,
    bootstrap_nsim: int = 500,
    p_val_thrs: float = 0.05
    ):
    
    """
    Compute forecast metrics.
    
    Parameters:
    -----------
    include_ref_metrics : bool, default False
        If True and ref_fcst is provided, also compute non-skill score metrics
        for the reference forecast. These will be named with '_ref' suffix.
    """
    
    METRICS = {
    "crps": lambda fcst, obs: xs.crps_ensemble(obs, fcst, dim=time_dim, member_dim=ensemble_dim),
    "crpsg": lambda fcst, obs: xs.crps_gaussian(obs, fcst.mean(ensemble_dim), fcst.mean(ensemble_dim)/fcst.mean(ensemble_dim), dim=time_dim),
    "rps": lambda fcst, obs: xs.rps(obs, fcst, dim=time_dim, member_dim=ensemble_dim, category_edges=rps_thresholds),
    "bs": lambda fcst, obs: xs.threshold_brier_score(obs, fcst, threshold=bs_threshold, dim=time_dim, member_dim=ensemble_dim),
    "rmse": lambda fcst, obs: xs.rmse(fcst.median(ensemble_dim), obs, dim=time_dim),
    "mae": lambda fcst, obs: xs.mae(fcst.median(ensemble_dim), obs, dim=time_dim),
    "mse": lambda fcst, obs: xs.mse(fcst.median(ensemble_dim), obs, dim=time_dim),
    "meanbias": lambda fcst, obs: (fcst.mean([ensemble_dim]) - obs).mean(time_dim),
    "medianbias": lambda fcst, obs: (fcst.median(ensemble_dim) - obs).mean(time_dim),
    "ensmeanpearsoncorr": lambda fcst, obs: xr.corr(obs, fcst.mean(ensemble_dim), dim=time_dim),
    "ensmedianpearsoncorr": lambda fcst, obs: xr.corr(obs, fcst.median(ensemble_dim), dim=time_dim),
    "ensmeanspearmancorr": lambda fcst, obs: xs.spearman_r(obs, fcst.mean(ensemble_dim), dim=time_dim),
    "ensmedianspearmancorr": lambda fcst, obs: xs.spearman_r(obs, fcst.median(ensemble_dim), dim=time_dim),
    "climbias": lambda fcst, obs: (fcst.mean([time_dim, ensemble_dim]) - obs.mean(time_dim))
}

    _METRICS = list(METRICS.keys()) + _SKILL_scores
    
    # Verify all requested metrics are available
    missing_metrics = set(metrics) - set(_METRICS)
    if missing_metrics:
        raise ValueError(f"The following metrics were not found: {missing_metrics}")

    # Compute all metrics
    METRICS_DS = []
    
    # Main forecast metrics
    for metric in metrics:
        METRICS_DS.extend(_compute_single_metric(metric, fcst, obs, METRICS,
                                                  suffix="", ref_fcst=ref_fcst,
                                                  time_dim=time_dim, bootstrap=bootstrap,
                                                  bootstrap_nsim=bootstrap_nsim, p_val_thrs=p_val_thrs))
    
    # Reference forecast metrics (only non-skill scores)
    if include_ref_metrics and ref_fcst is not None:
        ref_metrics_to_compute = [m for m in metrics if m not in _SKILL_scores]
        for metric in ref_metrics_to_compute:
            METRICS_DS.extend(
                _compute_single_metric(
                    metric, 
                    ref_fcst, obs, 
                    METRICS, 
                    "_" + ref_fcst.attrs["system"]
                )
            )
                              
    return xr.merge(METRICS_DS)

CBAR_KWARGS = {
    "shrink": 0.8, 
    "pad": 0.05,
    "orientation": "horizontal", #"label": None
    }

import matplotlib.colors as colors

def plot_metric(
    metric: xr.DataArray, 
    pval: xr.DataArray = None,
    ax: plt.Axes = None,
    mask: xr.DataArray = None,
    projection: ccrs.Projection = ccrs.PlateCarree(),
    add_colorbar: bool = True,
    cbar_kwargs: dict = CBAR_KWARGS, 
    subtitles_kwargs: dict = {"fontsize": 40},
    cbar_fontsize: int = 30,
    title: str = None,
    sign_thrsh: float = 0.05,
    **kwargs
    ):
    
    """
    Plot forecast metrics.
    """

    # ---------------------------
    #  CUSTOM COLORMAPS

    _auprc_cmap = plt.cm.YlOrRd
    _auprc_cmap.set_under(color='white')

    _crpss_cmap = plt.cm.YlOrRd
    _crpss_cmap.set_under(color='white')

    # Define the number of color segments for the lower and upper parts of the colormap
    n_segments = 28

    # Define the midpoint of the colormap
    midpoint = 0.5

    # Create the lower part of the colormap (blue to white)
    lower_colors = plt.cm.Blues_r(np.linspace(0.2, 1, n_segments))

    # Create the middle (white) part of the colormap
    n_white = 80 - 2 * n_segments  # Calculate the number of white segments
    white_colors = plt.cm.bwr(np.ones(n_white) * midpoint)

    # Create the upper part of the colormap (white to red)
    upper_colors = plt.cm.YlOrRd(np.linspace(0.2, 1, n_segments))

    # Combine all parts of the colormap
    combined_colors = np.vstack((lower_colors, white_colors, upper_colors))

    # Create the custom colormap
    _div_cmap = LinearSegmentedColormap.from_list('custom_diverging', combined_colors)

    def div_cmap():
        # Define the color segments
        colors = [
            (0.0, (0, 0, 0.5)),       # Dark blue
            (0.05, (0, 0, 0.75)),     # Medium blue
            (0.1, (0, 0.2, 1)),       # Blue
            (0.15, (0, 0.5, 1)),      # Light blue
            (0.2, (0.2, 0.7, 1)),     # Very light blue
            (0.3, (0.4, 0.8, 1)),     # Pale blue
            (0.4, (.9, .9, .9)),   # Light gray
            (0.6, (.9, .9, .9)),   # Light gray
            (0.7, (1, 1, 0)),         # Yellow
            (0.8, (1, 0.8, 0)),       # Light orange
            (0.85, (1, 0.6, 0)),      # Orange
            (0.9, (1, 0.4, 0)),       # Dark orange
            (0.95, (1, 0, 0)),        # Red
            (1.0, (0.5, 0, 0))        # Dark red
        ]
        
        # Create the colormap
        cmap = LinearSegmentedColormap.from_list("custom", colors, N=256)
        return cmap

    _div_cmap = div_cmap()

    _cmap = plt.cm.YlGnBu
    _cmap.set_under(color='pink')
    

    # Create a custom normalization with a midpoint at 0
    import matplotlib.colors as colors
    
    colors_undersea = plt.cm.RdPu_r(np.linspace(0, 0.9, 256))
    colors_land = plt.cm.YlGnBu(np.linspace(0, 1, 256))
    # whites should be between the values -0.05 and 0.05
    all_colors = np.vstack((colors_undersea, colors_land))
    terrain_map = colors.LinearSegmentedColormap.from_list(
    'terrain_map', all_colors)

    # make the norm:  Note the center is offset so that the land has more
    # dynamic range:
    divnorm = colors.BoundaryNorm(
        [-2, -1.5, -1, -0.5,-0.2, 0, 0.1, 0.2, 0.3, 0.4, 0.5],              
                    256)

    
    _args = {"scores": dict(cmap=terrain_map, extend="both", norm=divnorm, levels=22),
    #_args = {"scores": dict(cmap="PiYG", vmin=-.5, vmax=.5, levels=21, extend="both"),
             "bias": dict(cmap="bwr", vmin=-3, vmax=3, levels=21, extend="both"),
             "corr": dict(cmap=_div_cmap, vmin=-1, vmax=1, levels=21)
            }
    
    METRIC_KWARGS = {
        # -------------- Scores
        # ------------------------------------------------------------
        "crpss": _args["scores"], "crpsgs": _args["scores"], "rpss": _args["scores"],
        "bss": _args["scores"], "mses": _args["scores"], "rmses": _args["scores"], "maes": _args["scores"],
        # -------------- Scores
        # -------------------------------------------------------------
        "auprc": dict(cmap=_auprc_cmap, vmin=0.2, vmax=1, levels=11),
        "mae": dict(cmap="viridis", vmin=0, vmax=3, levels=21),
        "mse": dict(cmap="RdBu_r", vmin=0, vmax=3, levels=21),
        "rmse": dict(cmap="RdBu_r", vmin=0, vmax=3, levels=21),
        "mae": dict(cmap="viridis", vmin=0, vmax=3, levels=21),
        # -------------- Correlations
        # --------------------------------------------------------------
        "ensmeanspearmancorr": _args["corr"], "ensmedianspearmancorr": _args["corr"],
        "ensmeanpearsoncorr": _args["corr"], "ensmedianpearsoncorr": _args["corr"],
        # -------------- Biases
        # --------------------------------------------------------------
        "climbias": _args["bias"], "meanbias": _args["bias"], "medianbias": _args["bias"],
        # -------------- Forecasts
        # --------------------------------------------------------------
        "fcst": dict(cmap="bwr", vmin=-3, vmax=3, levels=21),
        "obs": dict(cmap="bwr", vmin=-3, vmax=3, levels=21),
        "fcstensmean": dict(cmap="bwr", vmin=-3, vmax=3, levels=21),
    }

    METRIC_TITLE = {

        "crpss": "CRPSS",
        "crpsgs": "CRPSG",
        "rpss": "RPSS",
        "bss": "BSS",
        "auprc_neg1sig": "AUPRC [-1$\sigma$]",
        "auprc_1sig": "AUPRC [1$\sigma$]",
        "auprc_neg2sig": "AUPRC [-2$\sigma$]",
        "auprc_2sig": "AUPRC [2$\sigma$]",
        #"ensmeancorr": "ens. mean corr.",
        #"ensmediancorr": "ens. median corr.",
        "mse": "MSE",
        "rmse": "RMSE",
        "rmses": "RMSS",
        "mses": "MSS",
        "maes": "MAS",
        "mae": "MAE",
        "ensmeanspearmancorr": "ens. mean Spearman corr.",
        "ensmedianspearmancorr": "ens. median Spearman corr.",
        "ensmeanpearsoncorr": "ens. mean Pearson corr.",
        "ensmedianpearsoncorr": "ens. median Pearson corr.",
        "climbias": "Clim. bias",
        "meanbias": "Mean bias",
        "medianbias": "Median bias",
        "fcst": "fcst",
        "obs": "obs",
        "fcstensmean": "fcst ens. mean"
        
    }

    _signi = xr.where(pval < sign_thrsh, 1, 0) if pval is not None else None
    ax, im = plot_array_to_map(
        metric, 
        mask=mask,
        ax=ax,
        projection=projection, 
        cbar_kwargs=cbar_kwargs,
        add_colorbar=add_colorbar,
        significance = _signi,
        fill_nan= -0.00001, #if metric.name.split("_")[0] in _SKILL_scores else None,
        **METRIC_KWARGS[metric.name.split("_")[0]],
        **kwargs
    )

    cbar = ax.collections[0].colorbar

    if add_colorbar:
        for l in cbar.ax.xaxis.get_ticklabels():
            l.set_weight("bold")
            l.set_fontsize(cbar_fontsize)

        # constrain the number of ticks labels plotted in the cbar to 5
        cbar.ax.locator_params(nbins=10)

    if title:

        plt.title(title, **subtitles_kwargs)
    else:
        plt.title("", **subtitles_kwargs)
        
    return ax, im, cbar